"""RasterDatasetDriver for zarr data."""

from copy import deepcopy
from functools import partial
from logging import Logger, getLogger
from os.path import splitext
from typing import Callable, ClassVar, List, Optional

import xarray as xr

from hydromt._typing import (
    Geom,
    SourceMetadata,
    StrPath,
    TimeRange,
    Variables,
    Zoom,
)
from hydromt._utils.unused_kwargs import _warn_on_unused_kwargs
from hydromt.data_catalog.drivers.preprocessing import PREPROCESSORS
from hydromt.data_catalog.drivers.raster.raster_dataset_driver import (
    RasterDatasetDriver,
)
from hydromt.error import NoDataStrategy, exec_nodata_strat

logger: Logger = getLogger(__name__)

_ZARR_EXT = ".zarr"
_NETCDF_EXT = [".nc", ".netcdf"]


class RasterDatasetXarrayDriver(RasterDatasetDriver):
    """
    Driver for RasterDataset using the xarray library: ``raster_xarray``.

    Supports reading and writing zarr and netcdf files using xarray.
    zarr files will be read using `xr.open_zarr` and netcdf files using
    `xr.open_mfdataset`.

    Driver **options** include:

    * preprocess: Optional[str], name of preprocessor to apply before merging datasets.
      Available preprocessors include: round_latlon, to_datetimeindex,
      remove_duplicates, harmonise_dims. See their docstrings for details.
    * ext_override: Optional[str], if set, will override the file extension check
      and try to read all files as the given extension. Useful when reading zarr
      files without the .zarr extension.
    * Any other option supported by `xr.open_zarr` or `xr.open_mfdataset`.

    """

    name = "raster_xarray"
    supports_writing = True
    SUPPORTED_EXTENSIONS: ClassVar[set[str]] = {_ZARR_EXT, *_NETCDF_EXT}

    def read(
        self,
        uris: List[str],
        *,
        mask: Optional[Geom] = None,
        variables: Optional[Variables] = None,
        time_range: Optional[TimeRange] = None,
        zoom: Optional[Zoom] = None,
        chunks: Optional[dict] = None,
        metadata: Optional[SourceMetadata] = None,
        handle_nodata: NoDataStrategy = NoDataStrategy.RAISE,
    ) -> xr.Dataset:
        """
        Read zarr data to an xarray DataSet.

        Args:
        """
        _warn_on_unused_kwargs(
            self.__class__.__name__,
            {
                "zoom": zoom,
            },
        )
        options = deepcopy(self.options)

        # Sort out the preprocessor
        preprocessor: Optional[Callable] = None
        preprocessor_name: Optional[str] = options.pop("preprocess", None)
        if preprocessor_name:
            preprocessor = PREPROCESSORS.get(preprocessor_name)
            if not preprocessor:
                raise ValueError(f"unknown preprocessor: '{preprocessor_name}'")

        # Check for the override flag
        ext_override: Optional[str] = options.pop("ext_override", None)
        if ext_override is not None:
            first_ext: str = ext_override
        else:
            first_ext: str = splitext(uris[0])[-1]

        # When is zarr, open like a zarr archive
        if first_ext == _ZARR_EXT:
            opn: Callable = partial(xr.open_zarr, **options)
            datasets = []
            for _uri in uris:
                ext = splitext(_uri)[-1]
                if ext != first_ext and not ext_override:
                    logger.warning(f"Reading zarr and {_uri} was not, skipping...")
                else:
                    datasets.append(
                        preprocessor(opn(_uri)) if preprocessor else opn(_uri)
                    )

            ds: xr.Dataset = xr.merge(datasets)

        # Normal netcdf file(s)
        elif first_ext in _NETCDF_EXT:
            filtered_uris = []
            for _uri in uris:
                ext = splitext(_uri)[-1]
                if ext != first_ext:
                    logger.warning(f"Reading netcdf and {_uri} was not, skipping...")
                else:
                    filtered_uris.append(_uri)

            ds: xr.Dataset = xr.open_mfdataset(
                filtered_uris, decode_coords="all", preprocess=preprocessor, **options
            )
        else:
            raise ValueError(
                f"Unknown extension for RasterDatasetXarrayDriver: {first_ext} "
            )
        for variable in ds.data_vars:
            if ds[variable].size == 0:
                exec_nodata_strat(
                    f"No data from driver: '{self.name}' for variable: '{variable}'",
                    strategy=handle_nodata,
                )
        return ds

    def write(self, path: StrPath, ds: xr.Dataset, **kwargs) -> str:
        """
        Write the RasterDataset to a local file using zarr.

        args:

        returns: str with written uri
        """
        no_ext, ext = splitext(path)
        # set filepath if incompat
        if ext not in self.SUPPORTED_EXTENSIONS:
            logger.warning(
                f"Unknown extension for RasterDatasetXarrayDriver: {ext},"
                "switching to zarr"
            )
            path = no_ext + _ZARR_EXT
            ext = _ZARR_EXT
        if ext == _ZARR_EXT:
            ds.to_zarr(path, mode="w", **kwargs)
        else:
            ds.to_netcdf(path, **kwargs)

        return str(path)



"""All the definitions of type aliases used in HydroMT."""

from datetime import datetime
from pathlib import Path
from typing import Any, Dict, List, Literal, Tuple, TypedDict, Union

import geopandas as gpd
from dateutil.parser import parse
from pydantic import ValidationInfo, ValidatorFunctionWrapHandler
from pydantic.functional_validators import (
    AfterValidator,
    BeforeValidator,
    WrapValidator,
)
from shapely.geometry.base import BaseGeometry
from typing_extensions import Annotated
from xarray import DataArray, Dataset

from hydromt._typing.model_mode import ModelMode


def _time_range_from_str(
    t: Tuple[Union[str, datetime], Union[str, datetime]],
) -> "TimeRange":
    if isinstance(t[0], str):
        t0 = parse(t[0])
    else:
        t0 = t[0]
    if isinstance(t[1], str):
        t1 = parse(t[1])
    else:
        t1 = t[1]
    return (t0, t1)


def _time_range_validate(tr: tuple[datetime, datetime]) -> tuple[datetime, datetime]:
    assert tr[0] >= tr[1], f"time range t0: '{tr[0]}' should be less than t1: '{tr[1]}'"
    return tr


def _validate_bbox(
    bbox: tuple[float, float, float, float],
) -> tuple[float, float, float, float]:
    assert bbox[0] < bbox[2], (
        f"bbox minx: '{bbox[0]}' should be less than maxx: '{bbox[2]}'."
    )
    assert bbox[1] < bbox[3], (
        f"bbox miny: '{bbox[1]}' should be less than maxy: '{bbox[3]}'."
    )
    return bbox


DataType = Literal[
    "DataFrame", "DataSet", "GeoDataFrame", "GeoDataSet", "RasterDataset"
]
Bbox = Annotated[Tuple[float, float, float, float], _validate_bbox]


def _validate_path(
    path: Any, handler: ValidatorFunctionWrapHandler, info: ValidationInfo
):
    if isinstance(path, str):
        path = Path(str)
    return handler(path, info)


Pathdantic = Annotated[Path, WrapValidator(_validate_path)]

StrPath = Union[str, Path]
GeoDataframeSource = StrPath
GeoDatasetSource = StrPath
RasterDatasetSource = StrPath
DatasetSource = StrPath

Crs = int
TotalBounds = Tuple[Bbox, Crs]
TimeRange = Annotated[
    Tuple[datetime, datetime],
    BeforeValidator(_time_range_from_str),
    AfterValidator(_time_range_validate),
]
Zoom = Union[int, Tuple[float, str]]  # level OR (resolution, unit)
Number = Union[int, float]
SourceSpecDict = TypedDict(
    "SourceSpecDict", {"source": str, "provider": str, "version": Union[str, int]}
)

XArrayDict = Dict[str, Union[DataArray, Dataset]]

ExportConfigDict = TypedDict(
    "ExportConfigDict",
    {"args": Dict[str, Any], "meta": Dict[str, Any], "sources": List[SourceSpecDict]},
)

Predicate = Literal[
    "intersects", "within", "contains", "overlaps", "crosses", "touches"
]

Geom = Union[gpd.GeoDataFrame, gpd.GeoSeries]
GpdShapeGeom = Union[Geom, BaseGeometry]

Data = Union[Dataset, DataArray]

Variables = Union[str, List[str]]

GeomBuffer = int

ModeLike = Union[ModelMode, str]



"""All drivers that can read RasterDatasets."""

from .raster_dataset_driver import RasterDatasetDriver
from .raster_xarray_driver import RasterDatasetXarrayDriver
from .rasterio_driver import RasterioDriver

__all__ = ["RasterDatasetDriver", "RasterDatasetXarrayDriver", "RasterioDriver"]



"""URIResolver using HydroMT naming conventions."""

from functools import reduce
from itertools import chain, product
from logging import Logger, getLogger
from typing import Any, Iterable, Optional

import pandas as pd
from fsspec.core import split_protocol

from hydromt._typing import (
    Geom,
    SourceMetadata,
    TimeRange,
    Zoom,
)
from hydromt._utils.naming_convention import _expand_uri_placeholders
from hydromt.data_catalog.uri_resolvers.uri_resolver import URIResolver
from hydromt.error import NoDataStrategy, exec_nodata_strat
from hydromt.gis.gis_utils import zoom_to_overview_level

logger: Logger = getLogger(__name__)


class ConventionResolver(URIResolver):
    """URIDataResolver using HydroMT naming conventions."""

    _uri_placeholders = frozenset(
        {"year", "month", "variable", "name", "overview_level"}
    )
    name = "convention"

    def _get_dates(
        self,
        keys: list[str],
        time_range: TimeRange,
    ) -> pd.PeriodIndex:
        """Obtain the dates the user is searching for."""
        t_range: pd.DatetimeIndex = pd.to_datetime(list(time_range))
        freq: str = "M" if "month" in keys else "Y"
        dates: pd.PeriodIndex = pd.period_range(*t_range, freq=freq)
        return dates

    def _resolve_wildcards(self, uris: Iterable[str]) -> set[str]:
        """Expand on the wildcards in the uris based on the filesystem."""

        def split_and_glob(uri: str) -> tuple[Optional[str], list[str]]:
            protocol, _ = split_protocol(uri)
            return (protocol, self.filesystem.glob(uri))

        def maybe_unstrip_protocol(
            pair: tuple[Optional[str], Iterable[str]],
        ) -> Iterable[str]:
            if pair[0] is not None:
                return map(
                    lambda uri: self.filesystem.unstrip_protocol(uri)
                    if not uri.startswith(pair[0])
                    else uri,
                    pair[1],
                )
            else:
                return pair[1]

        return set(
            chain.from_iterable(  # flatten result
                map(
                    lambda uri_pair: maybe_unstrip_protocol(
                        uri_pair
                    ),  # keep protocol in uri if present
                    reduce(
                        lambda uri_res, uri: uri_res + [split_and_glob(uri)], uris, []
                    ),
                )
            )
        )

    def resolve(
        self,
        uri: str,
        *,
        time_range: Optional[TimeRange] = None,
        mask: Optional[Geom] = None,
        zoom: Optional[Zoom] = None,
        variables: Optional[list[str]] = None,
        metadata: Optional[SourceMetadata] = None,
        handle_nodata: NoDataStrategy = NoDataStrategy.RAISE,
    ) -> list[str]:
        """Resolve the placeholders in the URI using naming conventions.

        Parameters
        ----------
        uri : str
            Unique Resource Identifier
        time_range : Optional[TimeRange], optional
            left-inclusive start end time of the data, by default None
        mask : Optional[Geom], optional
            A geometry defining the area of interest, by default None
        zoom: Optional[Zoom], optional
            zoom of the dataset, by default None
        variables : Optional[list[str]], optional
            Names of variables to return, or all if None, by default None
        metadata: Optional[SourceMetadata], optional
            DataSource metadata.
        handle_nodata : NoDataStrategy, optional
            how to react when no data is found, by default NoDataStrategy.RAISE

        Returns
        -------
        list[str]
            a list of expanded uris

        Raises
        ------
        NoDataException
            when no data is found and `handle_nodata` is `NoDataStrategy.RAISE`
        """
        logger.debug(f"attempting to resolve uri: {uri}")

        if metadata is None:
            metadata: SourceMetadata = SourceMetadata()

        uri_expanded, keys, _ = _expand_uri_placeholders(
            uri,
            placeholders=self._uri_placeholders,
            time_range=time_range,
            variables=variables,
        )
        if time_range:
            dates = self._get_dates(keys, time_range)
        else:
            dates = pd.PeriodIndex(["1970-01-01"], freq="d")  # fill any valid value
        if not variables:
            variables = [""]  # fill any valid value
        if zoom:
            zls_dict: Optional[dict[int, float]] = None
            if metadata is not None and hasattr(metadata, "zls_dict"):
                zls_dict = metadata.zls_dict
            crs = None if metadata is None else metadata.crs
            overview_level: int = zoom_to_overview_level(zoom, mask, zls_dict, crs) or 0
        else:
            overview_level = 0  # fill any valid value

        fmts: Iterable[dict[str, Any]] = map(
            lambda t: {
                "year": t[0].year,
                "month": t[0].month,
                "variable": t[1],
                "overview_level": overview_level,
            },
            product(dates, variables),
        )
        uris: list[str] = list(
            self._resolve_wildcards(map(lambda fmt: uri_expanded.format(**fmt), fmts))
        )
        if not uris:
            exec_nodata_strat(
                f"resolver '{self.name}' found no files.", strategy=handle_nodata
            )
            return []  # if ignore

        return uris



"""Data Adapters are generic for its HydroMT type and perform transformations."""

from .dataframe import DataFrameAdapter
from .dataset import DatasetAdapter
from .geodataframe import GeoDataFrameAdapter
from .geodataset import GeoDatasetAdapter
from .rasterdataset import RasterDatasetAdapter

__all__ = [
    "GeoDataFrameAdapter",
    "DatasetAdapter",
    "DataFrameAdapter",
    "GeoDatasetAdapter",
    "RasterDatasetAdapter",
]



"""Driver for GeoDataFrames."""

from abc import ABC, abstractmethod
from logging import Logger, getLogger
from typing import List, Optional

import geopandas as gpd

from hydromt._typing import SourceMetadata, StrPath
from hydromt.data_catalog.drivers import BaseDriver
from hydromt.error import NoDataStrategy

logger: Logger = getLogger(__name__)


class GeoDataFrameDriver(BaseDriver, ABC):
    """Abstract Driver to read GeoDataFrames."""

    @abstractmethod
    def read(
        self,
        uris: List[str],
        *,
        metadata: Optional[SourceMetadata] = None,
        handle_nodata: NoDataStrategy = NoDataStrategy.RAISE,
        **kwargs,
    ) -> gpd.GeoDataFrame:
        """Read in any compatible data source to a geopandas `GeoDataFrame`."""
        ...

    def write(
        self,
        path: StrPath,
        gdf: gpd.GeoDataFrame,
        **kwargs,
    ) -> str:
        """
        Write out a GeoDataFrame to file.

        Not all drivers should have a write function, so this method is not
        abstract.

        args:
        """
        raise NotImplementedError(
            f"Writing using driver '{self.name}' is not supported."
        )



"""GeoDatasetDriver for zarr data."""

from copy import copy
from functools import partial
from logging import Logger, getLogger
from os.path import splitext
from typing import Callable, ClassVar, List, Optional

import xarray as xr

from hydromt._typing import (
    Geom,
    SourceMetadata,
    StrPath,
    TimeRange,
)
from hydromt._typing.type_def import Predicate
from hydromt._utils.unused_kwargs import _warn_on_unused_kwargs
from hydromt.data_catalog.drivers.geodataset.geodataset_driver import GeoDatasetDriver
from hydromt.data_catalog.drivers.preprocessing import PREPROCESSORS
from hydromt.error import NoDataStrategy, exec_nodata_strat

logger: Logger = getLogger(__name__)

_ZARR_EXT = ".zarr"
_NETCDF_EXT = [".nc", ".netcdf"]


class GeoDatasetXarrayDriver(GeoDatasetDriver):
    """
    Driver for GeoDataset using the xarray library: ``geodataset_xarray``.

    Supports reading and writing zarr and netcdf files using xarray.
    zarr files will be read using `xr.open_zarr` and netcdf files using
    `xr.open_mfdataset`.

    Driver **options** include:

    * preprocess: Optional[str], name of preprocessor to apply before merging datasets.
      Available preprocessors include: round_latlon, to_datetimeindex,
      remove_duplicates, harmonise_dims. See their docstrings for details.
    * Any other option supported by `xr.open_zarr` or `xr.open_mfdataset`.

    """

    name: ClassVar[str] = "geodataset_xarray"
    supports_writing = True
    SUPPORTED_EXTENSIONS: ClassVar[set[str]] = {_ZARR_EXT, *_NETCDF_EXT}

    def read(
        self,
        uris: List[str],
        *,
        mask: Optional[Geom] = None,
        metadata: Optional[SourceMetadata] = None,
        predicate: Predicate = "intersects",
        time_range: Optional[TimeRange] = None,
        variables: Optional[List[str]] = None,
        handle_nodata: NoDataStrategy = NoDataStrategy.RAISE,
    ) -> xr.Dataset:
        """
        Read zarr data to an xarray DataSet.

        Args:
        """
        _warn_on_unused_kwargs(
            self.__class__.__name__,
            {
                "mask": mask,
                "time_range": time_range,
                "variables": variables,
                "predicate": predicate,
            },
        )
        options = copy(self.options)
        preprocessor: Optional[Callable] = None
        preprocessor_name: Optional[str] = options.pop("preprocess", None)
        if preprocessor_name:
            preprocessor = PREPROCESSORS.get(preprocessor_name)
            if not preprocessor:
                raise ValueError(f"unknown preprocessor: '{preprocessor_name}'")

        first_ext = splitext(uris[0])[-1]
        if first_ext == _ZARR_EXT:
            opn: Callable = partial(xr.open_zarr, **options)
            datasets = []
            for _uri in uris:
                ext = splitext(_uri)[-1]
                if ext != first_ext:
                    logger.warning(f"Reading zarr and {_uri} was not, skipping...")
                else:
                    datasets.append(
                        preprocessor(opn(_uri)) if preprocessor else opn(_uri)
                    )

            ds: xr.Dataset = xr.merge(datasets)
        elif first_ext in _NETCDF_EXT:
            filtered_uris = []
            for _uri in uris:
                ext = splitext(_uri)[-1]
                if ext != first_ext:
                    logger.warning(f"Reading netcdf and {_uri} was not, skipping...")
                else:
                    filtered_uris.append(_uri)

            ds: xr.Dataset = xr.open_mfdataset(
                filtered_uris, decode_coords="all", preprocess=preprocessor, **options
            )
        else:
            raise ValueError(
                f"Unknown extention for GeoDatasetXarrayDriver: {first_ext} "
            )
        for variable in ds.data_vars:
            if ds[variable].size == 0:
                exec_nodata_strat(
                    f"No data from driver: '{self.name}' for variable: '{variable}'",
                    strategy=handle_nodata,
                )
        return ds

    def write(self, path: StrPath, ds: xr.Dataset, **kwargs) -> str:
        """
        Write the GeoDataset to a local file using zarr.

        args:
        """
        ext = splitext(path)[-1]
        if ext == _ZARR_EXT:
            ds.vector.to_zarr(path, **kwargs)
        elif ext in _NETCDF_EXT:
            ds.vector.to_netcdf(path, **kwargs)
        else:
            raise ValueError(f"Unknown extension for GeoDatasetXarrayDriver: {ext} ")

        return path



# -*- coding: utf-8 -*-
"""command line interface for hydromt models."""

import logging
from ast import literal_eval
from datetime import datetime
from json import loads as json_decode
from os.path import join
from pathlib import Path
from typing import Any

import click
import numpy as np
from pydantic import ValidationError

from hydromt import __version__
from hydromt._typing.type_def import StrPath
from hydromt._utils import log
from hydromt._validators import Format
from hydromt._validators.data_catalog_v0x import DataCatalogV0Validator
from hydromt._validators.data_catalog_v1x import DataCatalogV1Validator
from hydromt._validators.model_config import HydromtModelSetup
from hydromt.cli import _utils
from hydromt.data_catalog import DataCatalog
from hydromt.error import NoDataStrategy
from hydromt.io import read_workflow_yaml, read_yaml, write_yaml
from hydromt.plugins import PLUGINS

logger = logging.getLogger(__name__)

HYDROMT_LOG_PATH = "hydromt.log"


def print_available_models(ctx, param, value):
    """Print the available models and exit.

    Parameters
    ----------
    ctx : click.Context
        The Click context object.
    value : bool
        The value of the parameter.
    """
    # param is required by click though we don't use it
    if not value:
        return {}
    click.echo(f"{PLUGINS.model_summary()}")
    ctx.exit()


def print_available_components(ctx, _param, value):
    """Print the available components and exit.

    Parameters
    ----------
    ctx : click.Context
        The Click context object.
    param : click.Parameter
        The Click parameter object.
    value : bool
        The value of the parameter.
    """
    if not value:
        return {}
    click.echo(PLUGINS.component_summary())
    ctx.exit()


def print_available_plugins(ctx, _param, value):
    """Print the available plugins and exit.

    Parameters
    ----------
    ctx : click.Context
        The Click context object.
    param : click.Parameter
        The Click parameter object.
    value : bool
        The value of the parameter.
    """
    if not value:
        return {}
    click.echo(f"{PLUGINS.plugin_summary()}")
    ctx.exit()


## common arguments / options
opt_config = click.option(
    "-i",
    "--config",
    type=click.Path(resolve_path=True),
    help="Path to hydroMT configuration file, for the model specific implementation.",
)
req_config = click.option(
    "-i",
    "--config",
    type=click.Path(resolve_path=True),
    help="Path to hydroMT configuration file, for the model specific implementation.",
    required=True,
)
export_dest_path = click.argument(
    "export_dest_path",
    type=click.Path(resolve_path=True, dir_okay=True, file_okay=False),
)
arg_root = click.argument(
    "MODEL_ROOT",
    type=click.Path(resolve_path=True, dir_okay=True, file_okay=False),
)

region_opt = click.option(
    "-r",
    "--region",
    type=str,
    default="{}",
    callback=_utils.parse_json,
    help="Set the region for which to build the model,"
    " e.g. {'subbasin': [-7.24, 62.09]}",
)

verbose_opt = click.option("--verbose", "-v", count=True, help="Increase verbosity.")

quiet_opt = click.option("--quiet", "-q", count=True, help="Decrease verbosity.")

data_opt = click.option(
    "-d",
    "--data",
    multiple=True,
    help="Path to local yaml data catalog file OR name of predefined data catalog.",
)

format_option = click.option(
    "--format",
)

deltares_data_opt = click.option(
    "--dd",
    "--deltares-data",
    is_flag=True,
    default=False,
    help='Flag: Shortcut to add the "deltares_data" catalog',
)

overwrite_opt = click.option(
    "--fo",
    "--force-overwrite",
    is_flag=True,
    default=False,
    help="Flag: If provided overwrite existing model files",
)
error_on_empty = click.option(
    "--error-on-empty",
    is_flag=True,
    default=False,
    help="Flag: Raise an error when attempting to export empty dataset instead of continuing",
)

cache_opt = click.option(
    "--cache",
    is_flag=True,
    default=False,
    help="Flag: If provided cache tiled rasterdatasets",
)

## MAIN


@click.group()
@click.version_option(__version__, message="HydroMT version: %(version)s")
@click.option(
    "--models",
    default=False,
    is_flag=True,
    is_eager=True,
    help="Print available model plugins and exit.",
    callback=print_available_models,
)
@click.option(
    "--components",
    default=False,
    is_flag=True,
    is_eager=True,
    help="Print available component plugins and exit.",
    callback=print_available_components,
)
@click.option(
    "--plugins",
    default=False,
    is_flag=True,
    is_eager=True,
    help="Print available component plugins and exit.",
    callback=print_available_plugins,
)
@click.pass_context
def main(ctx, models, components, plugins):
    """Command line interface for hydromt models."""
    if ctx.obj is None:
        ctx.obj = {}


## BUILD
@main.command(short_help="Build models")
@click.argument(
    "MODEL",
    type=str,
)
@arg_root
@req_config
@data_opt
@deltares_data_opt
@overwrite_opt
@cache_opt
@verbose_opt
@quiet_opt
@click.pass_context
def build(
    _ctx: click.Context,
    model,
    model_root,
    config,
    data,
    dd,
    fo,
    cache,
    verbose,
    quiet,
):
    """Build models from scratch.

    Example usage:
    --------------

    To build a wflow model:
    hydromt build wflow /path/to/model_root -i /path/to/wflow_config.yml
    -d deltares_data -d /path/to/data_catalog.yml -v

    To build a sfincs model:
    hydromt build sfincs /path/to/model_root  -i /path/to/sfincs_config.yml
    -d /path/to/data_catalog.yml -v
    """  # noqa: E501
    log_level = max(10, 30 - 10 * (verbose - quiet))
    log._setuplog(join(model_root, HYDROMT_LOG_PATH), log_level=log_level, append=False)
    logger.info(f"Building instance of {model} model at {model_root}.")
    logger.info("User settings:")
    modeltype, kwargs, steps = read_workflow_yaml(config, modeltype=model)

    # parse data catalog options from global section in config and cli options
    data_libs = np.atleast_1d(kwargs.pop("data_libs", [])).tolist()  # from global
    data_libs += list(data)  # add data catalogs from cli
    if dd and "deltares_data" not in data_libs:  # deltares_data from cli
        data_libs = ["deltares_data"] + data_libs  # prepend!
    try:
        # initialize model and create folder structure
        mode = "w+" if fo else "w"
        if modeltype not in PLUGINS.model_plugins:
            raise ValueError("Unknown model")
        mod = PLUGINS.model_plugins[modeltype](
            root=model_root,
            mode=mode,
            data_libs=data_libs,
            **kwargs,
        )
        mod.data_catalog.cache = cache
        # build model
        mod.build(steps=steps)

    except Exception as e:
        logger.exception(e)  # catch and log errors
        raise
    finally:
        log._wait_and_remove_file_handlers(logger)  # Release locks on logs


## UPDATE
@main.command(
    short_help="Update models",
)
@click.argument(
    "MODEL",
    type=str,
)
@arg_root
@click.option(
    "-o",
    "--model-out",
    type=click.Path(resolve_path=True, dir_okay=True, file_okay=False),
    help="Output model folder. Maps in MODEL_ROOT are overwritten if left empty.",
    default=None,
    callback=lambda c, p, v: v if v else c.params["model_root"],
)
@req_config
@data_opt
@deltares_data_opt
@overwrite_opt
@cache_opt
@quiet_opt
@verbose_opt
@click.pass_context
def update(
    _ctx: click.Context,
    model,
    model_root,
    model_out,
    config,
    data,
    dd,
    fo,
    cache,
    verbose,
    quiet,
):
    """Update a specific component of a model.

    Set an output directory to copy the edited model to a new folder, otherwise maps
    are overwritten.

    Example usage:
    --------------

    Update Wflow model components outlined in an .yml configuration file and
    write the model to a directory:
    hydromt update wflow /path/to/model_root  -o /path/to/model_out  -i /path/to/wflow_config.yml  -d /path/to/data_catalog.yml -v
    """  # noqa: E501
    # logger
    mode = "r+" if model_root == model_out else "r"
    log_level = max(10, 30 - 10 * (verbose - quiet))
    log._setuplog(join(model_out, HYDROMT_LOG_PATH), log_level=log_level)
    logger.info(f"Updating {model} model at {model_root} ({mode}).")
    logger.info(f"Output dir: {model_out}")
    # parse settings
    logger.info("User settings:")
    modeltype, kwargs, steps = read_workflow_yaml(config, modeltype=model)

    if modeltype not in PLUGINS.model_plugins:
        raise ValueError("Unknown model")
    # parse data catalog options from global section in config and cli options
    data_libs = np.atleast_1d(kwargs.pop("data_libs", [])).tolist()  # from global
    data_libs += list(data)  # add data catalogs from cli
    if dd and "deltares_data" not in data_libs:  # deltares_data from cli
        data_libs = ["deltares_data"] + data_libs  # prepend!
    try:
        # initialize model and create folder structure
        mod = PLUGINS.model_plugins[modeltype](
            root=model_root,
            mode=mode,
            data_libs=data_libs,
            **kwargs,
        )
        mod.data_catalog.cache = cache
        mod.update(model_out=model_out, steps=steps, forceful_overwrite=fo)
    except Exception as e:
        logger.exception(e)  # catch and log errors
        raise
    finally:
        log._wait_and_remove_file_handlers(logger)  # Release locks on logs


@main.command(
    short_help="Validate config / data catalog / region",
)
@click.option(
    "-m",
    "--model",
    type=str,
    default=None,
    help="Model name, e.g. wflow, sfincs, etc. to validate config file.",
)
@opt_config
@data_opt
@quiet_opt
@verbose_opt
@click.option("--format", type=click.Choice(list(Format)), default=Format.v1)
@click.option("--upgrade", is_flag=True)
@click.pass_context
def check(
    _ctx: click.Context,
    model: str | None,
    config,
    data,
    quiet: int,
    verbose: int,
    format: Format,
    upgrade: bool,
):
    """
    Verify that provided data catalog and config files are in the correct format.

    Additionally region bbox and geom can also be validated.

    Example usage:
    --------------

    Check data catalog file:
    hydromt check -d /path/to/data_catalog.yml -v

    Check data catalog and grid_model config file:
    hydromt check -m grid_model -d /path/to/data_catalog.yml -i /path/to/model_config.yml -v

    With region:
    hydromt check -m grid_model -d /path/to/data_catalog.yml -i /path/to/model_config.yml -r '{'bbox': [-1,-1,1,1]}' -v

    """  # noqa: E501
    # logger
    log_level = max(10, 30 - 10 * (verbose - quiet))
    log._setuplog(join(".", HYDROMT_LOG_PATH), log_level=log_level)
    failed = False
    try:
        for cat_path in data:
            logger.info(f"Validating catalog at {cat_path}")
            try:
                if format == Format.v0:
                    v0_catalog = DataCatalogV0Validator.from_yml(cat_path)
                    if upgrade:
                        v1_catalog = DataCatalogV1Validator.from_v0(v0_catalog)

                        p = Path(cat_path)
                        write_yaml(
                            p.with_stem(f"{p.stem}_v1"),
                            v1_catalog.model_dump(
                                exclude_unset=True,
                                exclude_defaults=True,
                                exclude_none=True,
                            ),
                        )

                # if we add more versions don't forget to update this
                # statement here
                else:
                    DataCatalogV1Validator.from_yml(cat_path)

                logger.info(f"Catalog {cat_path} is valid!")
            except ValidationError as e:
                failed = True
                msg = f"Catalog {cat_path} has the following error(s): {e}"
                logger.error(msg)

        if config:
            logger.info(f"Validating config at {config}")
            if format == Format.v0:
                logger.error(
                    "Because of the dynamic nature of workflow files, v0.x files cannot be "
                    "checked by hydromt v1. Skipping..."
                )
            else:
                try:
                    modeltype, kwargs, steps = read_workflow_yaml(
                        config, modeltype=model
                    )
                    config_dict = {
                        "modeltype": modeltype,
                        "global": kwargs,
                        "steps": steps,
                    }
                    HydromtModelSetup(**config_dict)
                    logger.info("Model config valid!")

                except (ValidationError, ValueError) as e:
                    failed = True
                    msg = f"Workflow yaml {config} has the following error(s): {e}"
                    logger.error(msg)

        if failed:
            # We've already presented the errors above
            # now we simply raise without extra msg to fail the process
            # so it has the correct exit code
            raise ValueError(
                "hydromt check command failed. See error messages above for the details"
            )

    except Exception as e:
        logger.exception(e)  # catch and log errors
        raise
    finally:
        log._wait_and_remove_file_handlers(logger)  # Release locks on logs


## Export
@main.command(
    short_help="Export data",
)
@click.option(
    "-s",
    "--source",
    multiple=True,
    help="Name of the data source to export.",
)
@click.option(
    "-t",
    "--time-range",
    help="Time tuple as a list of two strings, e.g. ['2010-01-01', '2022-12-31']",
)
@click.option(
    "-b",
    "--bbox",
    help="a bbox in EPSG:4236 designating the region of which to export the data",
)
@region_opt
@export_dest_path
@opt_config
@data_opt
@deltares_data_opt
@overwrite_opt
@error_on_empty
@quiet_opt
@verbose_opt
@click.pass_context
def export(
    _ctx: click.Context,
    export_dest_path: Path,
    source: str | None,
    time_range: str | None,
    bbox: tuple[float, float, float, float] | None,
    config: Path | None,
    region: dict[Any, Any] | None,
    data: list[StrPath] | None,
    dd: bool,
    fo: bool,
    error_on_empty: bool,
    quiet: int,
    verbose: int,
):
    """Export the data from a catalog.

    Example usage:
    --------------

    export the data of a single data source, in a particular region, for a particular time range
    hydromt export -r "{'bbox': [4.6891,52.9750,4.9576,53.1994]}" -s era5_hourly -d ../hydromt/data/catalogs/artifact_data.yml -t '["2010-01-01", "2022-12-31"]' path/to/output_dir

    export a single data source from the deltares data catalog without time/space slicing
    hydromt export -d deltares_data -s era5_hourly path/to/output_dir

    export data as detailed in an export config yaml file
    hydromt export -i /path/to/export_config.yaml path/to/output_dir
    """  # noqa: E501
    # logger
    log_level = max(10, 30 - 10 * (verbose - quiet))
    log._setuplog(join(export_dest_path, HYDROMT_LOG_PATH), log_level=log_level)
    logger.info(f"Output dir: {export_dest_path}")

    if error_on_empty:
        handle_nodata = NoDataStrategy.RAISE
    else:
        handle_nodata = NoDataStrategy.IGNORE

    data_libs: list[StrPath] = list(data) if data else []
    if dd and "deltares_data" not in data_libs:  # deltares_data from cli
        data_libs.insert(0, "deltares_data")

    sources: list[str] = []
    if source:
        if isinstance(source, str):
            sources = [source]
        else:
            sources = list(source)

    # these need to be defined even if config does not exist
    unit_conversion = True
    meta = {}
    append = False

    if config:
        config_dict = read_yaml(config)["export_data"]
        if "data_libs" in config_dict.keys():
            data_libs = data_libs + config_dict.pop("data_libs")
        time_range = config_dict.pop("time_range", None)
        region = region or config_dict.pop("region", None)
        if isinstance(region, str):
            region = json_decode(region)

        sources = sources + config_dict["sources"]

        unit_conversion = config_dict.pop("unit_conversion", True)
        meta = config_dict.pop("meta", {})
        append = config_dict.pop("append", False)

    data_catalog = DataCatalog(data_libs=data_libs)
    _ = data_catalog.sources  # initialise lazy loading

    if time_range:
        if isinstance(time_range, str):
            tup = literal_eval(time_range)
        else:
            tup = time_range
        time_start = datetime.strptime(tup[0], "%Y-%m-%d")
        time_end = datetime.strptime(tup[1], "%Y-%m-%d")
        time_tup = (time_start, time_end)
    else:
        time_tup = None

    if isinstance(bbox, str):
        bbox = literal_eval(bbox)

    try:
        data_catalog.export_data(
            export_dest_path,
            source_names=sources,
            bbox=bbox,
            time_range=time_tup,
            unit_conversion=unit_conversion,
            metadata=meta,
            append=append,
            handle_nodata=handle_nodata,
            force_overwrite=fo,
        )

    except Exception as e:
        logger.exception(e)  # catch and log errors
        raise
    finally:
        log._wait_and_remove_file_handlers(logger)  # Release locks on logs


if __name__ == "__main__":
    main()



"""Runtime Settings for HydroMT."""

from pathlib import Path

from pydantic import Field
from pydantic_settings import BaseSettings

from hydromt._typing import Pathdantic


class Settings(BaseSettings):
    """Runtime Settings for HydroMT."""

    cache_root: Pathdantic = Field(default=Path.home() / ".hydromt")


SETTINGS = Settings()



"""All drivers for readin GeoDataframe datasets."""

from .geodataframe_driver import GeoDataFrameDriver
from .pyogrio_driver import PyogrioDriver
from .table_driver import GeoDataFrameTableDriver

__all__ = [
    "GeoDataFrameDriver",
    "PyogrioDriver",
    "GeoDataFrameTableDriver",
]



"""Driver using rasterio for RasterDataset."""

import copy
from logging import Logger, getLogger
from pathlib import Path
from typing import Any, ClassVar, Dict, List, Optional, Tuple, Union

import numpy as np
import rasterio
import rasterio.errors
import xarray as xr
from pyproj import CRS

from hydromt._typing import (
    Geom,
    SourceMetadata,
    StrPath,
    TimeRange,
    Variables,
    Zoom,
)
from hydromt._utils.caching import _cache_vrt_tiles
from hydromt._utils.temp_env import temp_env
from hydromt._utils.uris import _strip_scheme
from hydromt.config import SETTINGS
from hydromt.data_catalog.drivers.raster.raster_dataset_driver import (
    RasterDatasetDriver,
)
from hydromt.error import NoDataStrategy, exec_nodata_strat
from hydromt.gis.gis_utils import zoom_to_overview_level
from hydromt.io.readers import open_mfraster

logger: Logger = getLogger(__name__)


class RasterioDriver(RasterDatasetDriver):
    """
    Driver for RasterDataset using the rasterio library: ``rasterio``.

    Supports reading and writing raster files using rasterio.

    Driver **options** include:

    * mosaic: bool, if True and multiple uris are given, will mosaic the datasets
      together using `rasterio.merge.merge`. Default is False.
    * mosaic_kwargs: dict, additional keyword arguments to pass to
      `rasterio.merge.merge`.
    * cache: bool, if True and reading from VRT files, will cache the tiles
      locally to speed up reading. Default is False.
    * cache_root: str, root directory for caching. Default is taken from
      `hydromt.config.SETTINGS.cache_root`.
    * cache_dir: str, subdirectory for caching. Default is the stem of the first
      uri without extension.
    * Any other option supported by `hydromt.io.readers.open_mfraster`.

    """

    name = "rasterio"
    SUPPORTED_EXTENSIONS: ClassVar[set[str]] = {
        "." + extension
        for extension in rasterio.drivers.raster_driver_extensions()
        if extension != "nc"
    }  # Exclude netcdf as a supported file type

    def read(
        self,
        uris: List[str],
        *,
        mask: Optional[Geom] = None,
        time_range: Optional[TimeRange] = None,
        variables: Optional[Variables] = None,
        zoom: Optional[Zoom] = None,
        chunks: Optional[dict] = None,
        metadata: Optional[SourceMetadata] = None,
        handle_nodata: NoDataStrategy = NoDataStrategy.RAISE,
    ) -> xr.Dataset:
        """Read data using rasterio."""
        if metadata is None:
            metadata = SourceMetadata()
        # build up kwargs for open_raster
        options = copy.deepcopy(self.options)
        mosaic_kwargs: Dict[str, Any] = self.options.get("mosaic_kwargs", {})
        mosaic: bool = options.pop("mosaic", False) and len(uris) > 1

        # get source-specific options
        cache_root: str = str(
            options.pop("cache_root", SETTINGS.cache_root),
        )

        # Check for caching, default to false
        cache_flag = options.pop("cache", False)

        # Caching portion, only when the flag is True and the file format is vrt
        if all([uri.endswith(".vrt") for uri in uris]) and cache_flag:
            cache_dir = Path(cache_root) / options.pop(
                "cache_dir",
                Path(
                    _strip_scheme(uris[0])[1]
                ).stem,  # default to first uri without extension
            )
            uris_cached = []
            for uri in uris:
                cached_uri: str = _cache_vrt_tiles(
                    uri, geom=mask, fs=self.filesystem, cache_dir=cache_dir
                )
                uris_cached.append(cached_uri)
            uris = uris_cached

        if mask is not None:
            mosaic_kwargs.update({"mask": mask})

        # get mosaic kwargs
        if mosaic_kwargs:
            options.update({"mosaic_kwargs": mosaic_kwargs})

        if np.issubdtype(type(metadata.nodata), np.number):
            options.update(nodata=metadata.nodata)

        # Fix overview level
        if zoom:
            try:
                zls_dict: Dict[int, float] = metadata.zls_dict
                crs: Optional[CRS] = metadata.crs
            except AttributeError:  # pydantic extra=allow on SourceMetadata
                zls_dict, crs = self._get_zoom_levels_and_crs(uris[0])

            overview_level: Optional[int] = zoom_to_overview_level(
                zoom, mask, zls_dict, crs
            )
            if overview_level:
                # NOTE: overview levels start at zoom_level 1, see _get_zoom_levels_and_crs
                options.update(overview_level=overview_level - 1)

        if chunks is not None:
            options.update({"chunks": chunks})

        # If the metadata resolver has already resolved the overview level,
        # trying to open zoom levels here will result in an error.
        # Better would be to separate uriresolver and driver: https://github.com/Deltares/hydromt/issues/1023
        # Then we can implement looking for a overview level in the driver.
        def _open() -> Union[xr.DataArray, xr.Dataset]:
            try:
                return open_mfraster(uris, mosaic=mosaic, **options)
            except rasterio.errors.RasterioIOError as e:
                if "Cannot open overview level" in str(e):
                    options.pop("overview_level")
                    return open_mfraster(uris, mosaic=mosaic, **options)
                else:
                    raise

        # rasterio uses specific environment variable for s3 access.
        try:
            anon: str = self.filesystem.anon
        except AttributeError:
            anon: str = ""

        if anon:
            with temp_env(**{"AWS_NO_SIGN_REQUEST": "true"}):
                ds = _open()
        else:
            ds = _open()

        # Mosaic's can mess up the chunking, which can error during writing
        # Or maybe setting
        chunks = options.get("chunks")
        if chunks is not None:
            ds = ds.chunk(chunks=chunks)

        # rename ds with single band if single variable is requested
        if variables is not None and len(variables) == 1 and len(ds.data_vars) == 1:
            ds = ds.rename({list(ds.data_vars.keys())[0]: list(variables)[0]})

        for variable in ds.data_vars:
            if ds[variable].size == 0:
                exec_nodata_strat(
                    f"No data from driver: '{self.name}' for variable: '{variable}'",
                    strategy=handle_nodata,
                )
        return ds

    def write(self, path: StrPath, ds: xr.Dataset, **kwargs) -> str:
        """Write out a RasterDataset using rasterio."""
        raise NotImplementedError()

    @staticmethod
    def _get_zoom_levels_and_crs(uri: str) -> Tuple[Dict[int, float], int]:
        """Get zoom levels and crs from adapter or detect from tif file if missing."""
        zoom_levels = {}
        crs = None
        try:
            with rasterio.open(uri) as src:
                res = abs(src.res[0])
                crs = src.crs
                overviews = [src.overviews(i) for i in src.indexes]
                if len(overviews[0]) > 0:  # check overviews for band 0
                    # check if identical
                    if not all([o == overviews[0] for o in overviews]):
                        raise ValueError("Overviews are not identical across bands")
                    # dict with overview level and corresponding resolution
                    zls = [1] + overviews[0]
                    zoom_levels = {i: res * zl for i, zl in enumerate(zls)}
        except rasterio.RasterioIOError as e:
            logger.warning(f"IO error while detecting zoom levels: {e}")
        return zoom_levels, crs



from typing import Any, Dict, Iterable

from fsspec import AbstractFileSystem, filesystem
from pydantic import PlainSerializer, PlainValidator
from typing_extensions import Annotated


def validate_filesystem(
    _fs: Any,
) -> AbstractFileSystem:
    if isinstance(_fs, str):
        # _fs is protocol
        return filesystem(_fs)
    elif isinstance(_fs, dict):
        # _fs is dict with build args for fsspec filesystem.
        if protocol := _fs.pop("protocol", None):
            # fsspec to_dict gets some args
            args: Iterable[Any] = _fs.pop("args", [])
            return filesystem(protocol, *args, **_fs)
        else:
            raise ValueError(f"Filesystem dict {_fs} requires 'protocol'.")
    elif isinstance(_fs, AbstractFileSystem):
        # _fs is already deserialized
        return _fs
    else:
        raise ValueError(f"Unknown filesystem: {_fs}")


def serialize_filesystem(_fs: AbstractFileSystem) -> Dict[str, Any]:
    fs_dict: Dict[str, str] = _fs.to_dict(
        include_password=False
    )  # do not include any passwords.
    # cls is not required
    fs_dict.pop("cls")
    if not fs_dict.get("args"):
        fs_dict.pop("args")
    return fs_dict


FS = Annotated[
    AbstractFileSystem,
    PlainValidator(validate_filesystem),
    PlainSerializer(serialize_filesystem),
]



"""Driver for reading in GeoDataFrames from tabular formats."""

from copy import deepcopy
from logging import Logger, getLogger
from typing import ClassVar, List, Optional

import geopandas as gpd

from hydromt._typing.metadata import SourceMetadata
from hydromt.data_catalog.drivers.geodataframe.geodataframe_driver import (
    GeoDataFrameDriver,
)
from hydromt.error import NoDataStrategy, exec_nodata_strat
from hydromt.io.readers import open_vector_from_table

logger: Logger = getLogger(__name__)

# possible labels for the x and y dimensions
X_DIM_LABELS = ("x", "longitude", "lon", "long")
Y_DIM_LABELS = ("y", "latitude", "lat")


class GeoDataFrameTableDriver(GeoDataFrameDriver):
    """
    Driver for GeoDataFrame from tabular formats: ``geodataframe_table``.

    Supports reading point geometries from csv, excel (xls, xlsx), and parquet files
    using a combination of the pandas and geopandas libraries.

    Driver **options** include:

    * x_dim: Optional[str], name of the column containing the x coordinate. Not needed
      if one of the default names is used ('x', 'longitude', 'lon', 'long').
    * y_dim: Optional[str], name of the column containing the y coordinate. Not needed
      if one of the default names is used ('y', 'latitude', 'lat').
    * Any other option supported by the underlying pandas read functions,
      e.g. `pd.read_csv`, `pd.read_excel`, `pd.read_parquet`.

    """

    name: ClassVar[str] = "geodataframe_table"
    SUPPORTED_EXTENSIONS: ClassVar[set[str]] = {".csv", ".xlsx", ".xls", ".parquet"}

    def read(
        self,
        uris: List[str],
        *,
        metadata: Optional[SourceMetadata] = None,
        handle_nodata: NoDataStrategy = NoDataStrategy.RAISE,
        **kwargs,
    ) -> gpd.GeoDataFrame:
        """Read tabular data using a combination of the pandas and geopandas libraries."""
        if not metadata:
            metadata = SourceMetadata()
        if len(uris) > 1:
            raise ValueError(
                "DataFrame: Reading multiple files with the "
                f"{self.__class__.__name__} driver is not supported."
            )

        _uri: str = uris[0]

        options = deepcopy(self.options)
        x_dim = options.pop("x_dim", None)
        y_dim = options.pop("y_dim", None)
        gdf = open_vector_from_table(
            path=_uri,
            x_dim=x_dim,
            y_dim=y_dim,
            crs=metadata.crs,
            **options,
        )
        if gdf.index.size == 0:
            exec_nodata_strat(
                f"No data from driver {self}'.",
                strategy=handle_nodata,
            )
        return gdf



"""Driver for handling IO of GeoDatasets."""

from abc import ABC, abstractmethod
from logging import getLogger
from typing import List, Optional

import xarray as xr

from hydromt._typing import Geom, SourceMetadata, StrPath, TimeRange
from hydromt._typing.type_def import Predicate
from hydromt.data_catalog.drivers.base_driver import BaseDriver
from hydromt.error import NoDataStrategy

logger = getLogger(__name__)


class GeoDatasetDriver(BaseDriver, ABC):
    """Abstract Driver to read GeoDatasets."""

    @abstractmethod
    def read(
        self,
        uris: List[str],
        *,
        mask: Optional[Geom] = None,
        predicate: Predicate = "intersects",
        variables: Optional[List[str]] = None,
        time_range: Optional[TimeRange] = None,
        metadata: Optional[SourceMetadata] = None,
        handle_nodata: NoDataStrategy = NoDataStrategy.RAISE,
        **kwargs,
    ) -> Optional[xr.Dataset]:
        """
        Read in any compatible data source to an xarray Dataset.

        args:
        """
        ...

    def write(
        self,
        path: StrPath,
        ds: xr.Dataset,
        **kwargs,
    ) -> str:
        """
        Write out a GeoDataset to file.

        Not all drivers should have a write function, so this method is not
        abstract.

        args:
        """
        raise NotImplementedError(
            f"Writing using driver '{self.name}' is not supported."
        )



from typing import Any, List, Optional, Union

from pydantic import PlainSerializer, PlainValidator
from pyproj import CRS as ProjCRS
from pyproj.exceptions import ProjError
from typing_extensions import Annotated


def _serialize_crs(crs: ProjCRS) -> Union[int, List[str], str]:
    # Try epsg authority first
    epsg: Optional[int] = crs.to_epsg()
    if not epsg:
        # Then try any authority
        auth: Optional[List[str, str]] = list(
            crs.to_authority()
        )  # cast to list to be serializable to yaml
        if not auth:
            # reserve wkt for last
            return crs.to_wkt()
        return auth
    return epsg


def _validate_crs(crs: Any) -> ProjCRS:
    try:
        return ProjCRS.from_user_input(crs)
    except ProjError:
        return ProjCRS.from_authority("ESRI", crs)  # fallback on ESRI


CRS = Annotated[ProjCRS, PlainValidator(_validate_crs), PlainSerializer(_serialize_crs)]



"""All drivers for reading GeoDatasets."""

from .geodataset_driver import GeoDatasetDriver
from .vector_driver import GeoDatasetVectorDriver
from .xarray_driver import GeoDatasetXarrayDriver

__all__ = ["GeoDatasetDriver", "GeoDatasetVectorDriver", "GeoDatasetXarrayDriver"]



"""Implementations related to logging."""

import os
import sys
from logging import (
    FileHandler,
    Formatter,
    Logger,
    StreamHandler,
    captureWarnings,
    getLogger,
)
from typing import Optional

from hydromt import __version__

FMT = "%(asctime)s - %(name)s - %(module)s - %(levelname)s - %(message)s"

__all__ = ["_setuplog"]


def _wait_and_remove_file_handlers(logger: Logger):
    for handler in logger.handlers:
        if isinstance(handler, FileHandler):
            # remove handler first, as otherwise new calls to the logger may open the
            # same filename again.
            logger.removeHandler(handler)
            # wait for all messages to be processed
            handler.flush()
            # then close the handler
            handler.close()


def _setuplog(
    path: Optional[str] = None,
    log_level: int = 20,
    fmt: str = FMT,
    append: bool = True,
):
    """Set up the logging on sys.stdout and file if path is given.

    Parameters
    ----------
    path : str, optional
        path to logfile, by default None
    log_level : int, optional
        Log level [0-50], by default 20 (info)
    fmt : str, optional
        log message formatter
    append : bool, optional
        Whether to append (True) or overwrite (False) to a logfile at path,
        by default True
    """
    main_logger: Logger = getLogger("hydromt")
    _wait_and_remove_file_handlers(main_logger)
    captureWarnings(True)
    main_logger.setLevel(log_level)
    console = StreamHandler(sys.stdout)
    console.setLevel(log_level)
    console.setFormatter(Formatter(fmt))
    main_logger.addHandler(console)
    if path is not None:
        if append is False and os.path.isfile(path):
            os.unlink(path)
        _add_filehandler(main_logger, path, log_level=log_level, fmt=fmt)
    main_logger.info(f"HydroMT version: {__version__}")


def _add_filehandler(logger, path, log_level=20, fmt=FMT):
    """Add file handler to logger."""
    if not os.path.isdir(os.path.dirname(path)):
        os.makedirs(os.path.dirname(path))
    isfile = os.path.isfile(path)
    ch = FileHandler(path)
    ch.setFormatter(Formatter(fmt))
    ch.setLevel(log_level)
    logger.addHandler(ch)
    if isfile:
        logger.debug(f"Appending log messages to file {path}.")
    else:
        logger.debug(f"Writing log messages to new file {path}.")



"""Utility functions for hydromt that have no other home."""

from .caching import _cache_vrt_tiles, _copy_to_local
from .dataset import (
    _rename_vars,
    _set_metadata,
    _shift_dataset_time,
    _single_var_as_array,
    _slice_temporal_dimension,
)
from .deep_merge import _deep_merge
from .dictionaries import _partition_dictionaries
from .elevation import _elevation2rgba, _rgba2elevation
from .log import _setuplog
from .nodata import _has_no_data, _set_raster_nodata, _set_vector_nodata
from .path import _make_config_paths_abs, _make_config_paths_relative
from .rgetattr import _rgetattr
from .steps_validator import _validate_steps
from .unused_kwargs import _warn_on_unused_kwargs
from .uris import _is_valid_url, _strip_scheme

__all__ = [
    "_cache_vrt_tiles",
    "_copy_to_local",
    "_deep_merge",
    "_rename_vars",
    "_set_metadata",
    "_setuplog",
    "_shift_dataset_time",
    "_single_var_as_array",
    "_slice_temporal_dimension",
    "_partition_dictionaries",
    "_elevation2rgba",
    "_rgba2elevation",
    "_has_no_data",
    "_set_raster_nodata",
    "_set_vector_nodata",
    "_rgetattr",
    "_validate_steps",
    "_warn_on_unused_kwargs",
    "_is_valid_url",
    "_strip_scheme",
    "_make_config_paths_abs",
    "_make_config_paths_relative",
]


class _classproperty(property):
    def __get__(self, owner_self, owner_cls):
        return self.fget(owner_cls)



"""Drivers responsible for reading and writing data."""

from .base_driver import BaseDriver
from .dataframe import DataFrameDriver, PandasDriver
from .dataset import DatasetDriver, DatasetXarrayDriver
from .geodataframe import GeoDataFrameDriver, GeoDataFrameTableDriver, PyogrioDriver
from .geodataset import GeoDatasetDriver, GeoDatasetVectorDriver, GeoDatasetXarrayDriver
from .preprocessing import (
    harmonise_dims,
    remove_duplicates,
    round_latlon,
    to_datetimeindex,
)
from .raster import RasterDatasetDriver, RasterDatasetXarrayDriver, RasterioDriver

__all__ = [
    "BaseDriver",
    "DatasetDriver",
    "DatasetXarrayDriver",
    "DataFrameDriver",
    "GeoDataFrameDriver",
    "GeoDataFrameTableDriver",
    "GeoDatasetDriver",
    "GeoDatasetVectorDriver",
    "GeoDatasetXarrayDriver",
    "PandasDriver",
    "PyogrioDriver",
    "RasterDatasetDriver",
    "RasterDatasetXarrayDriver",
    "RasterioDriver",
    "harmonise_dims",
    "remove_duplicates",
    "round_latlon",
    "to_datetimeindex",
]

# define hydromt driver entry points
# see also hydromt.driver group in pyproject.toml
__hydromt_eps__ = [
    "DatasetXarrayDriver",
    "GeoDataFrameTableDriver",
    "GeoDatasetVectorDriver",
    "GeoDatasetXarrayDriver",
    "PandasDriver",
    "PyogrioDriver",
    "RasterDatasetXarrayDriver",
    "RasterioDriver",
]



"""Driver to read geodataframes using Pyogrio."""

from logging import Logger, getLogger
from os.path import splitext
from typing import ClassVar, List, Optional, Union

import geopandas as gpd
import pandas as pd
from pyogrio import read_dataframe, read_info, write_dataframe
from pyproj import CRS

from hydromt._typing import Bbox, Geom, StrPath
from hydromt.data_catalog.drivers.geodataframe.geodataframe_driver import (
    GeoDataFrameDriver,
)
from hydromt.error import NoDataStrategy, exec_nodata_strat

logger: Logger = getLogger(__name__)


class PyogrioDriver(GeoDataFrameDriver):
    """
    Driver for GeoDataFrame using the pyogrio library: ``pyogrio``.

    Supports reading and writing files supported by the OGR library,
    including geopackage, shapefile, geojson and flatgeobuf.

    Options in this driver are passed to `pyogrio.read_dataframe`.
    """

    name = "pyogrio"
    supports_writing = True
    SUPPORTED_EXTENSIONS: ClassVar[set[str]] = {".gpkg", ".shp", ".geojson", ".fgb"}

    def read(
        self,
        uris: List[str],
        *,
        mask: Optional[Geom] = None,
        variables: Optional[List[str]] = None,
        handle_nodata: NoDataStrategy = NoDataStrategy.RAISE,
        **kwargs,
    ) -> gpd.GeoDataFrame:
        """
        Read data using pyogrio.

        Warning
        -------
        The 'predicate' and 'metadata' keyword arguments are unused in this method and
        are only present in the method's signature for compatibility with other
        functions.
        """
        if len(uris) > 1:
            raise ValueError(
                "DataFrame: Reading multiple files with the "
                f"{self.__class__.__name__} driver is not supported."
            )
        elif len(uris) == 0:
            gdf = gpd.GeoDataFrame()
        else:
            _uri = uris[0]
            if mask is not None:
                bbox = _bbox_from_file_and_mask(_uri, mask=mask)
            else:
                bbox = None
            gdf: Union[pd.DataFrame, gpd.GeoDataFrame] = read_dataframe(
                _uri, bbox=bbox, columns=variables, **self.options
            )
        if not isinstance(gdf, gpd.GeoDataFrame):
            raise IOError(f"DataFrame from uri: '{_uri}' contains no geometry column.")

        if gdf.index.size == 0:
            exec_nodata_strat(f"No data from driver {self}'.", strategy=handle_nodata)
        return gdf

    def write(
        self,
        path: StrPath,
        gdf: gpd.GeoDataFrame,
        **kwargs,
    ) -> str:
        """
        Write out a GeoDataFrame to file using pyogrio.

        args:
        """
        no_ext, ext = splitext(path)
        if ext not in self.SUPPORTED_EXTENSIONS:
            logger.warning(
                f"driver {self.name} has no support for extension {ext}"
                "switching to .fgb."
            )
            path = no_ext + ".fgb"

        write_dataframe(gdf, path, **kwargs)

        return str(path)


def _bbox_from_file_and_mask(
    uri: str,
    mask: Geom,
) -> Optional[Bbox]:
    """Create a bbox from the file metadata and mask given.

    Pyogrio's mask or bbox arguments require a mask or bbox in the same CRS as the data.
    This function takes the mask filter and crs of the input data
    and returns a bbox in the same crs as the data based on the input filters.

    Parameters
    ----------
    uri: str,
        URI of the data.
    mask: GeoDataFrame | GeoSeries | BaseGeometry
        mask to filter the data while reading.
    """
    source_crs = None
    if source_crs_str := read_info(uri).get("crs"):
        source_crs = CRS.from_user_input(source_crs_str)

    if not source_crs:
        logger.warning(
            f"Reading from uri: '{uri}' without CRS definition. Filtering with crs:"
            f" {mask.crs}, cannot compare crs."
        )
    elif mask.crs != source_crs:
        mask = mask.to_crs(source_crs)

    return tuple(mask.total_bounds)



"""DataSource class for the GeoDataset type."""

from logging import Logger, getLogger
from os.path import basename, splitext
from typing import Any, ClassVar, Dict, List, Literal, Optional, Union, cast

import pandas as pd
import xarray as xr
from fsspec import filesystem
from pydantic import Field
from pyproj import CRS
from pyproj.exceptions import CRSError
from pystac import Asset as StacAsset
from pystac import Catalog as StacCatalog
from pystac import Item as StacItem
from pystac import MediaType

from hydromt._typing import (
    Bbox,
    Geom,
    StrPath,
    TimeRange,
    TotalBounds,
)
from hydromt._typing.type_def import GeomBuffer, Predicate
from hydromt.data_catalog.adapters.geodataset import GeoDatasetAdapter
from hydromt.data_catalog.drivers.geodataset.geodataset_driver import GeoDatasetDriver
from hydromt.data_catalog.sources.data_source import DataSource
from hydromt.error import NoDataStrategy
from hydromt.gis.gis_utils import _parse_geom_bbox_buffer

logger: Logger = getLogger(__name__)


class GeoDatasetSource(DataSource):
    """DataSource class for the GeoDatasetSource type."""

    data_type: ClassVar[Literal["GeoDataset"]] = "GeoDataset"
    _fallback_driver_read: ClassVar[str] = "geodataset_vector"
    _fallback_driver_write: ClassVar[str] = "geodataset_xarray"

    driver: GeoDatasetDriver
    data_adapter: GeoDatasetAdapter = Field(default_factory=GeoDatasetAdapter)

    def read_data(
        self,
        *,
        mask: Optional[Geom] = None,
        predicate: Predicate = "intersects",
        variables: Optional[List[str]] = None,
        time_range: Optional[TimeRange] = None,
        single_var_as_array: bool = True,
        handle_nodata: NoDataStrategy = NoDataStrategy.RAISE,
    ) -> Optional[Union[xr.Dataset, xr.DataArray]]:
        """
        Read data from this source.

        Args:
        """
        self._mark_as_used()
        self._log_start_read_data()

        # Transform time_range and variables to match the data source
        tr = self.data_adapter._to_source_timerange(time_range)
        vrs = self.data_adapter._to_source_variables(variables)

        uris: List[str] = self.uri_resolver.resolve(
            self.full_uri,
            time_range=tr,
            variables=vrs,
            metadata=self.metadata,
            handle_nodata=handle_nodata,
        )

        ds: Optional[xr.Dataset] = self.driver.read(
            uris,
            time_range=tr,
            variables=vrs,
            metadata=self.metadata,
            handle_nodata=handle_nodata,
        )
        return self.data_adapter.transform(
            ds,
            self.metadata,
            predicate=predicate,
            single_var_as_array=single_var_as_array,
            mask=mask,
            variables=variables,
            time_range=time_range,
            handle_nodata=handle_nodata,
        )

    def to_file(
        self,
        file_path: StrPath,
        *,
        driver_override: Optional[GeoDatasetDriver] = None,
        bbox: Optional[Bbox] = None,
        mask: Optional[Geom] = None,
        buffer: GeomBuffer = 0,
        predicate: Predicate = "intersects",
        variables: Optional[List[str]] = None,
        time_range: Optional[TimeRange] = None,
        single_var_as_array: bool = True,
        handle_nodata: NoDataStrategy = NoDataStrategy.RAISE,
        **kwargs,
    ) -> Optional["GeoDatasetSource"]:
        """
        Write the GeoDatasetSource to a local file.

        args:
        """
        if driver_override is None and not self.driver.supports_writing:
            # default to fallback driver.
            driver: GeoDatasetDriver = GeoDatasetDriver.model_validate(
                self._fallback_driver_write
            )
        elif driver_override:
            if not driver_override.supports_writing:
                raise RuntimeError(
                    f"driver: '{driver_override.name}' does not support writing data."
                )
            driver: GeoDatasetDriver = driver_override
        else:
            # use local filesystem
            driver: GeoDatasetDriver = self.driver.model_copy(
                update={"filesystem": filesystem("local")}
            )

        if bbox is not None or (mask is not None and buffer > 0):
            mask = _parse_geom_bbox_buffer(mask, bbox, buffer)
        ds: Optional[Union[xr.Dataset, xr.DataArray]] = self.read_data(
            mask=mask,
            predicate=predicate,
            variables=variables,
            single_var_as_array=single_var_as_array,
            time_range=time_range,
            handle_nodata=handle_nodata,
        )
        if ds is None:  # handle_nodata == ignore
            return None

        dest_path: str = driver.write(
            file_path,
            ds,
            **kwargs,
        )

        # update driver based on local path
        update: Dict[str, Any] = {"uri": dest_path, "root": None, "driver": driver}

        return self.model_copy(update=update)

    def get_bbox(self, crs: Optional[CRS] = None, detect: bool = True) -> TotalBounds:
        """Return the bounding box and espg code of the dataset.

        if the bounding box is not set and detect is True,
        :py:meth:`hydromt.GeoDatasetAdapter.detect_bbox` will be used to detect it.

        Parameters
        ----------
        detect: bool, Optional
            whether to detect the bounding box if it is not set. If False, and it's not
            set None will be returned.

        Returns
        -------
        bbox: Tuple[np.float64,np.float64,np.float64,np.float64]
            the bounding box coordinates of the data. coordinates are returned as
            [xmin,ymin,xmax,ymax]
        crs: int
            The ESPG code of the CRS of the coordinates returned in bbox
        """
        bbox = self.metadata.extent.get("bbox", None)
        crs = cast(int, crs)
        if bbox is None and detect:
            bbox, crs = self.detect_bbox()

        return bbox, crs

    def get_time_range(
        self,
        detect: bool = True,
    ) -> TimeRange:
        """Detect the time range of the dataset.

        if the time range is not set and detect is True,
        :py:meth:`hydromt.GeoDatasetAdapter.detect_time_range` will be used
        to detect it.


        Parameters
        ----------
        detect: bool, Optional
            whether to detect the time range if it is not set. If False, and it's not
            set None will be returned.

        Returns
        -------
        range: Tuple[np.datetime64, np.datetime64]
            A tuple containing the start and end of the time dimension. Range is
            inclusive on both sides.
        """
        time_range = self.metadata.extent.get("time_range", None)
        if time_range is None and detect:
            time_range = self.detect_time_range()

        return time_range

    def detect_bbox(
        self,
        ds: Optional[xr.Dataset] = None,
    ) -> TotalBounds:
        """Detect the bounding box and crs of the dataset.

        If no dataset is provided, it will be fetched according to the settings in the
        adapter. also see :py:meth:`hydromt.GeoDatasetAdapter.get_data`. the
        coordinates are in the CRS of the dataset itself, which is also returned
        alongside the coordinates.


        Parameters
        ----------
        ds: xr.Dataset, xr.DataArray, Optional
            the dataset to detect the bounding box of.
            If none is provided, :py:meth:`hydromt.GeoDatasetAdapter.get_data`
            will be used to fetch the it before detecting.

        Returns
        -------
        bbox: Tuple[np.float64,np.float64,np.float64,np.float64]
            the bounding box coordinates of the data. coordinates are returned as
            [xmin,ymin,xmax,ymax]
        crs: int
            The ESPG code of the CRS of the coordinates returned in bbox
        """
        if ds is None:
            ds = self.read_data()
        crs = ds.vector.crs.to_epsg()
        bounds = ds.vector.bounds

        return bounds, crs

    def detect_time_range(
        self,
        ds: Optional[xr.Dataset] = None,
    ) -> TimeRange:
        """Detect the temporal range of the dataset.

        If no dataset is provided, it will be fetched accodring to the settings in the
        addapter. also see :py:meth:`hydromt.GeoDatasetAdapter.get_data`.

        Parameters
        ----------
        ds: xr.Dataset, xr.DataArray, Optional
            the dataset to detect the time range of. It must have a time dimentsion set.
            If none is provided, :py:meth:`hydromt.GeoDatasetAdapter.get_data`
            will be used to fetch the it before detecting.

        Returns
        -------
        range: Tuple[np.datetime64, np.datetime64]
            A tuple containing the start and end of the time dimension. Range is
            inclusive on both sides.
        """
        if ds is None:
            ds = self.read_data()
        return (
            ds[ds.vector.time_dim].min().values,
            ds[ds.vector.time_dim].max().values,
        )

    def to_stac_catalog(
        self,
        handle_nodata: NoDataStrategy = NoDataStrategy.IGNORE,
    ) -> Optional[StacCatalog]:
        """
        Convert a geodataset into a STAC Catalog representation.

        The collection will contain an asset for each of the associated files.


        Parameters
        ----------
        - handle_nodata (str, optional): The error handling strategy.
          Options are: "raise" to raise an error on failure, "IGNORE" to skip
          the dataset on failure

        Returns
        -------
        - Optional[StacCatalog]: The STAC Catalog representation of the dataset, or
          None if the dataset was skipped.
        """
        try:
            bbox, crs = self.get_bbox(detect=True)
            bbox = list(bbox)
            start_dt, end_dt = self.get_time_range(detect=True)
            start_dt = pd.to_datetime(start_dt)
            end_dt = pd.to_datetime(end_dt)
            props = {**self.metadata.model_dump(exclude_none=True), "crs": crs}
            ext = splitext(self.uri)[-1]
            if ext in [".nc", ".vrt"]:
                media_type = MediaType.HDF5
            else:
                raise RuntimeError(
                    f"Unknown extension: {ext} cannot determine media type"
                )
        except (IndexError, KeyError, CRSError) as e:
            if handle_nodata == NoDataStrategy.IGNORE:
                logger.warning(
                    "Skipping {name} during stac conversion because"
                    "because detecting spacial extent failed."
                )
                return
            else:
                raise e

        stac_catalog = StacCatalog(
            self.name,
            description=self.name,
        )
        stac_item = StacItem(
            self.name,
            geometry=None,
            bbox=bbox,
            properties=props,
            datetime=None,
            start_datetime=start_dt,
            end_datetime=end_dt,
        )
        stac_asset = StacAsset(str(self.uri), media_type=media_type)
        base_name = basename(self.uri)
        stac_item.add_asset(base_name, stac_asset)

        stac_catalog.add_item(stac_item)
        return stac_catalog

    @classmethod
    def _infer_default_driver(
        cls,
        uri: str | None = None,
    ) -> str:
        return super()._infer_default_driver(uri, GeoDatasetDriver)



"""Implementation for the RasterDatasetAdapter."""

from __future__ import annotations

from logging import getLogger
from typing import Any, Dict, List, Optional, Union, cast

import numpy as np
import pyproj
import xarray as xr
from pyproj import CRS

from hydromt._typing import (
    Data,
    Geom,
    SourceMetadata,
    TimeRange,
    Variables,
)
from hydromt._utils import (
    _has_no_data,
    _shift_dataset_time,
    _single_var_as_array,
    _slice_temporal_dimension,
)
from hydromt.data_catalog.adapters.data_adapter_base import DataAdapterBase
from hydromt.error import NoDataException, NoDataStrategy, exec_nodata_strat
from hydromt.gis.raster import GEO_MAP_COORD
from hydromt.gis.raster_utils import _meridian_offset

logger = getLogger(__name__)

__all__ = ["RasterDatasetAdapter"]


class RasterDatasetAdapter(DataAdapterBase):
    """Implementation for the RasterDatasetAdapter."""

    def transform(
        self,
        ds: xr.Dataset,
        metadata: SourceMetadata,
        *,
        mask: Optional[Geom] = None,
        align: Optional[bool] = None,
        variables: Optional[Variables] = None,
        time_range: Optional[TimeRange] = None,
        handle_nodata: NoDataStrategy = NoDataStrategy.RAISE,
        single_var_as_array: bool = True,
        buffer: int = 0,
    ) -> Union[xr.Dataset, xr.DataArray, None]:
        """Filter and harmonize the input RasterDataset.

        Parameters
        ----------
        ds : xr.Dataset
            input RasterDataset
        metadata : SourceMetadata
            source metadata
        mask : Optional[gpd.GeoDataFrame], optional
            mask to filter by geometry, by default None
        variables : Optional[List[str]], optional
            variable filter, by default None
        time_range : Optional[TimeRange], optional
            filter start and end times, by default None
        handle_nodata : NoDataStrategy, optional
            how to handle no data being present in the result, by default NoDataStrategy.RAISE
        single_var_as_array : bool, optional
            whether to return a xr.DataArray if only a single variable is present, by default True
        buffer : int, optional
            Buffer around the `bbox` or `geom` area of interest expressed in resolution multiplicity,
            by default 0

        Returns
        -------
        Optional[xr.Dataset]
            The filtered and harmonized RasterDataset, or None if no data was available

        Raises
        ------
        ValueError
            if not all variables are found in the data
        NoDataException
            if no data in left after slicing and handle_nodata is NoDataStrategy.RAISE
        """
        try:
            if _has_no_data(ds):
                raise NoDataException()
            # rename variables and parse data and attrs
            ds = self._rename_vars(ds)
            ds = self._validate_spatial_dims(ds)
            ds = self._set_crs(ds, metadata.crs)
            ds = self._set_nodata(ds, metadata)
            ds = _shift_dataset_time(dt=self.unit_add.get("time", 0), ds=ds)
            # slice data
            ds = RasterDatasetAdapter._slice_data(
                ds,
                variables,
                mask,
                align,
                time_range,
                buffer=buffer,
            )
            if _has_no_data(ds):
                raise NoDataException()

            # uniformize data
            ds = self._apply_unit_conversions(ds)
            ds = self._set_metadata(ds, metadata)
            # return array if single var and single_var_as_array
            return _single_var_as_array(ds, single_var_as_array, variables)
        except NoDataException:
            exec_nodata_strat(
                "No data was read from source",
                strategy=handle_nodata,
            )

    def _rename_vars(self, ds: Data) -> Data:
        rm = {k: v for k, v in self.rename.items() if k in ds}
        ds = ds.rename(rm)
        return ds

    def _validate_spatial_dims(self, ds: Data) -> Data:
        if GEO_MAP_COORD in ds.data_vars:
            ds = ds.set_coords(GEO_MAP_COORD)
        try:
            ds.raster.set_spatial_dims()
            # transpose dims to get y and x dim last
            x_dim = ds.raster.x_dim
            y_dim = ds.raster.y_dim
            ds = ds.transpose(..., y_dim, x_dim)
        except ValueError:
            raise ValueError(
                f"RasterDataset: No valid spatial coords found in data {self.path}"
            )
        return ds

    def _set_crs(self, ds: Data, crs: Optional[CRS]) -> Data:
        # set crs
        if ds.raster.crs is None and crs is not None:
            ds.raster.set_crs(crs)
        elif ds.raster.crs is None:
            raise ValueError("RasterDataset: CRS not defined in data catalog or data.")
        elif crs is not None and ds.raster.crs != pyproj.CRS.from_user_input(crs):
            logger.warning(
                "RasterDataset: CRS from data catalog does not match CRS "
                " of data. The original CRS will be used. Please check your catalog."
            )
        return ds

    @staticmethod
    def _slice_data(
        ds: Data,
        variables: Optional[List] = None,
        mask: Optional[Geom] = None,
        align: Optional[float] = None,
        time_range: Optional[TimeRange] = None,
        buffer: int = 0,
    ) -> Optional[xr.Dataset]:
        """Filter the RasterDataset.

        Parameters
        ----------
        ds : xr.Dataset
            input Dataset.
        variables : Optional[List[str]], optional
            variable filter, by default None
        mask : Optional[gpd.GeoDataFrame], optional
            mask to filter by geometry, by default None
        align : Optional[float], optional
            resolution to align the bounding box, by default None
        time_range : Optional[TimeRange], optional
            filter start and end times, by default None
        buffer : int, optional
            Buffer around the `bbox` or `geom` area of interest expressed in resolution multiplicity,
            by default 0

        Returns
        -------
        Optional[Union[xr.Dataset, xr.DataArray]]
            The filtered and harmonized RasterDataset, or None if no data was available

        Raises
        ------
        ValueError
            if not all variables are found in the data
        NoDataException
            if no data in left after slicing and handle_nodata is NoDataStrategy.RAISE
        """
        if isinstance(ds, xr.DataArray):  # xr.DataArray has no variables
            if ds.name is None:
                # dummy name, required to create dataset
                # renamed to variable in _single_var_as_array
                ds.name = "data"
            ds = ds.to_dataset()
        elif variables is not None:  # xr.Dataset has variables
            variables = cast(List, np.atleast_1d(variables).tolist())
            if len(variables) > 1 or len(ds.data_vars) > 1:
                mvars = [var not in ds.data_vars for var in variables]
                if any(mvars):
                    raise NoDataException(f"RasterDataset: variables not found {mvars}")
                ds = ds[variables]
        if time_range is not None:
            ds = _slice_temporal_dimension(ds, time_range)
        if mask is not None:
            ds = RasterDatasetAdapter._slice_spatial_dimensions(
                ds, mask, buffer=buffer, align=align
            )

        if _has_no_data(ds):
            return None
        else:
            return ds

    @staticmethod
    def _slice_spatial_dimensions(
        ds: Data,
        mask: Optional[Geom] = None,
        align: Optional[float] = None,
        buffer: int = 0,
    ):
        # make sure bbox is in data crs
        bbox = None
        crs = ds.raster.crs
        epsg = crs.to_epsg()  # this could return None
        if mask is not None:
            bbox = mask.to_crs(crs).total_bounds
        # work with 4326 data that is defined at 0-360 degrees longtitude
        w, _, e, _ = ds.raster.bounds
        if epsg == 4326 and np.isclose(e - w, 360):  # allow for rounding errors
            ds = _meridian_offset(ds, bbox)

        # clip with bbox
        if bbox is not None:
            # check if bbox is fully covered
            bbox_str = ", ".join([f"{c:.3f}" for c in bbox])

            def _lt_or_close(a: float, b: float) -> bool:
                return np.isclose(a, b) or a < b

            w, s, e, n = ds.raster.bounds

            if not any(
                map(_lt_or_close, (w, s, bbox[2], bbox[3]), (bbox[0], bbox[1], e, n))
            ):
                logger.warning(
                    f"Dataset [{w}, {s}, {e}, {n}] does not fully cover bbox [{bbox_str}]"
                )

            logger.debug(f"Clip to [{bbox_str}] (epsg:{epsg}))")
            ds = ds.raster.clip_bbox(bbox, align=align, buffer=buffer)

        if _has_no_data(ds):
            return None
        else:
            return ds

    def _apply_unit_conversions(self, ds: Data):
        unit_names = list(self.unit_mult.keys()) + list(self.unit_add.keys())
        unit_names = [k for k in unit_names if k in ds.data_vars]
        if len(unit_names) > 0:
            logger.debug(f"Convert units for {len(unit_names)} variables.")
        for name in list(set(unit_names)):  # unique
            m = self.unit_mult.get(name, 1)
            a = self.unit_add.get(name, 0)
            da = ds[name]
            attrs = da.attrs.copy()
            nodata_isnan = da.raster.nodata is None or np.isnan(da.raster.nodata)
            # nodata value is explicitly set to NaN in case no nodata value is provided
            nodata = np.nan if nodata_isnan else da.raster.nodata
            data_bool = ~np.isnan(da) if nodata_isnan else da != nodata
            ds[name] = xr.where(data_bool, da * m + a, nodata)
            ds[name].attrs.update(attrs)  # set original attributes
            ds[name].raster.set_nodata(nodata)  # reset nodata in case of change

        return ds

    def _set_nodata(self, ds, metadata: SourceMetadata):
        """Parse and apply nodata values from the data catalog."""
        # set nodata value
        if nodata := metadata.nodata is not None:
            if not isinstance(nodata, dict):
                no_data_values: Dict[str, Any] = {
                    k: metadata.nodata for k in ds.data_vars.keys()
                }
            else:
                no_data_values: Dict[str, Any] = nodata
            for k in ds.data_vars:
                mv = no_data_values.get(k, None)
                if mv is not None and ds[k].raster.nodata is None:
                    ds[k].raster.set_nodata(mv)
        return ds

    def _set_metadata(self, ds, metadata: SourceMetadata):
        # unit attributes
        attrs = metadata.attrs
        for k in attrs:
            ds[k].attrs.update(attrs[k])
        # set meta data
        ds.attrs.update(metadata.model_dump(exclude=["attrs"], exclude_unset=True))
        return ds



"""URIResolvers obtain multiple URIs before being passed to Drivers."""

from .convention_resolver import ConventionResolver
from .raster_tindex_resolver import RasterTindexResolver
from .uri_resolver import URIResolver

__all__ = [
    "ConventionResolver",
    "URIResolver",
    "RasterTindexResolver",
]

# define hydromt uri resolver entry points
# see also hydromt.uri_resolver group in pyproject.toml
__hydromt_eps__ = [
    "ConventionResolver",
    "RasterTindexResolver",
]



"""URIResolver for raster tindex files."""

from logging import Logger, getLogger
from os.path import abspath, dirname, join
from pathlib import Path
from typing import List, Optional, Union

import geopandas as gpd

from hydromt._typing import (
    SourceMetadata,
    TimeRange,
    Zoom,
)
from hydromt.data_catalog.uri_resolvers.uri_resolver import URIResolver
from hydromt.error import NoDataStrategy, exec_nodata_strat

logger: Logger = getLogger(__name__)


class RasterTindexResolver(URIResolver):
    """Implementation of the URIResolver for raster tindex files."""

    name = "raster_tindex"

    def resolve(
        self,
        uri: str,
        *,
        time_range: Optional[TimeRange] = None,
        zoom: Optional[Zoom] = None,
        mask: Optional[gpd.GeoDataFrame] = None,
        variables: Union[int, tuple[float, str], None] = None,
        metadata: Optional[SourceMetadata] = None,
        handle_nodata: NoDataStrategy = NoDataStrategy.RAISE,
    ) -> List[str]:
        """Resolve URIs of a raster tindex file.

        Parameters
        ----------
        uri : str
            Unique Resource Identifier
        time_range : Optional[TimeRange], optional
            left-inclusive start end time of the data, by default None
        mask : Optional[Geom], optional
            A geometry defining the area of interest, by default None
        zoom: Optional[Zoom], optional
            zoom_level of the dataset, by default None
        variables : Optional[List[str]], optional
            Names of variables to return, or all if None, by default None
        metadata: Optional[SourceMetadata], optional
            DataSource metadata.
        handle_nodata : NoDataStrategy, optional
            how to react when no data is found, by default NoDataStrategy.RAISE

        Returns
        -------
        List[str]
            a list of expanded uris

        Raises
        ------
        NoDataException
            when no data is found and `handle_nodata` is `NoDataStrategy.RAISE`
        """
        logger.debug(f"attempting to resolve uri: {uri}")

        if mask is None:
            raise ValueError(f"Resolver {self.name} needs a mask")
        gdf = gpd.read_file(uri)
        gdf = gdf.iloc[gdf.sindex.query(mask.to_crs(gdf.crs).union_all())]
        tileindex: Optional[str] = self.options.get("tileindex")
        if tileindex is None:
            raise ValueError(
                f"{self.__class__.__name__} needs options specifying 'tileindex'"
            )
        if gdf.index.size == 0:
            exec_nodata_strat(
                f"resolver '{self.name}' found no intersecting tiles.",
                strategy=handle_nodata,
            )
            return []  # in case of ignore

        elif tileindex not in gdf.columns:
            raise IOError(
                f'Tile index "{tileindex}" column missing in tile index file.'
            )
        else:
            root = dirname(uri)
            paths = []
            for p in gdf[tileindex]:
                path = Path(str(p))
                if not path.is_absolute():
                    paths.append(str(Path(abspath(join(root, p)))))
        return paths



"""DataSources responsible for validating the DataCatalog."""

# importing all data sources for discovery, factory needs to be imported last.
from .data_source import DataSource  # noqa: I001
from .dataframe import DataFrameSource
from .dataset import DatasetSource
from .geodataframe import GeoDataFrameSource
from .geodataset import GeoDatasetSource
from .rasterdataset import RasterDatasetSource

from .factory import create_source  # noqa: I001

__all__ = [
    "DataSource",
    "DataFrameSource",
    "DatasetSource",
    "GeoDataFrameSource",
    "GeoDatasetSource",
    "RasterDatasetSource",
    "create_source",
]



"""DatasetDriver for zarr data."""

from copy import copy
from functools import partial
from logging import Logger, getLogger
from os.path import splitext
from typing import Callable, ClassVar, List, Optional

import xarray as xr

from hydromt._typing import (
    SourceMetadata,
    StrPath,
    TimeRange,
)
from hydromt._utils.unused_kwargs import _warn_on_unused_kwargs
from hydromt.data_catalog.drivers.dataset.dataset_driver import DatasetDriver
from hydromt.data_catalog.drivers.preprocessing import PREPROCESSORS
from hydromt.error import NoDataStrategy, exec_nodata_strat

logger: Logger = getLogger(__name__)


class DatasetXarrayDriver(DatasetDriver):
    """
    Driver for Dataset using xarray: ``dataset_xarray``.

    Supports reading and writing zarr and netcdf files using xarray.
    zarr files will be read using `xr.open_zarr` and netcdf files using
    `xr.open_mfdataset`.

    Driver **options** include:

    * preprocess: Optional[str], name of preprocessor to apply before merging datasets.
      Available preprocessors include: round_latlon, to_datetimeindex,
      remove_duplicates, harmonise_dims. See their docstrings for details.
    * Any other option supported by `xr.open_zarr` or `xr.open_mfdataset`.

    """

    name: ClassVar[str] = "dataset_xarray"
    supports_writing = True

    def read(
        self,
        uris: List[str],
        *,
        metadata: Optional[SourceMetadata] = None,
        time_range: Optional[TimeRange] = None,
        variables: Optional[List[str]] = None,
        handle_nodata: NoDataStrategy = NoDataStrategy.RAISE,
    ) -> xr.Dataset:
        """
        Read zarr data to an xarray DataSet.

        Args:
        """
        _warn_on_unused_kwargs(
            self.__class__.__name__,
            {
                "time_range": time_range,
                "variables": variables,
                "metadata": metadata,
            },
        )
        options = copy(self.options)
        preprocessor: Optional[Callable] = None
        preprocessor_name: Optional[str] = options.pop("preprocess", None)
        if preprocessor_name:
            preprocessor = PREPROCESSORS.get(preprocessor_name)
            if not preprocessor:
                raise ValueError(f"unknown preprocessor: '{preprocessor_name}'")

        first_ext = splitext(uris[0])[-1]
        if first_ext == ".zarr":
            opn: Callable = partial(xr.open_zarr, **options)
            datasets = []
            for _uri in uris:
                ext = splitext(_uri)[-1]
                if ext != first_ext:
                    logger.warning(f"Reading zarr and {_uri} was not, skipping...")
                else:
                    datasets.append(
                        preprocessor(opn(_uri)) if preprocessor else opn(_uri)
                    )

            ds: xr.Dataset = xr.merge(datasets)
        elif first_ext in [".nc", ".netcdf"]:
            filtered_uris = []
            for _uri in uris:
                ext = splitext(_uri)[-1]
                if ext != first_ext:
                    logger.warning(f"Reading netcdf and {_uri} was not, skipping...")
                else:
                    filtered_uris.append(_uri)

            ds: xr.Dataset = xr.open_mfdataset(
                filtered_uris, decode_coords="all", preprocess=preprocessor, **options
            )
        else:
            raise ValueError(f"Unknown extension for DatasetXarrayDriver: {first_ext} ")
        for variable in ds.data_vars:
            if ds[variable].size == 0:
                exec_nodata_strat(
                    f"No data from driver: '{self.name}' for variable: '{variable}'",
                    strategy=handle_nodata,
                )
        return ds

    def write(self, path: StrPath, ds: xr.Dataset, **kwargs) -> str:
        """
        Write the Dataset to a local file using zarr.

        args:
        """
        ext = splitext(path)[-1]
        if ext == ".zarr":
            ds.to_zarr(path, **kwargs)
        elif ext in [".nc", ".netcdf"]:
            ds.to_netcdf(path, **kwargs)
        else:
            raise ValueError(f"Unknown extension for DatasetXarrayDriver: {ext} ")

        return str(path)



"""Abstract DataSource class."""

from __future__ import annotations

from abc import ABC
from copy import deepcopy
from logging import Logger, getLogger
from os.path import abspath, join, splitext
from pathlib import Path, PurePath
from typing import Any, ClassVar, Dict, List, Optional, TypeVar, Union

from pydantic import (
    BaseModel,
    ConfigDict,
    Field,
    PrivateAttr,
    SerializerFunctionWrapHandler,
    model_serializer,
    model_validator,
)

from hydromt._typing import DataType, SourceMetadata
from hydromt._utils.uris import _is_valid_url
from hydromt.data_catalog.adapters.data_adapter_base import DataAdapterBase
from hydromt.data_catalog.drivers import BaseDriver
from hydromt.data_catalog.uri_resolvers import ConventionResolver, URIResolver
from hydromt.error import NoDataException, NoDataStrategy

logger: Logger = getLogger(__name__)

T = TypeVar("T")


class DataSource(BaseModel, ABC):
    """
    A DataSource is a parsed section of a DataCatalog.

    The DataSource, specific for a data type within HydroMT, is responsible for
    validating the input from the DataCatalog, to
    ensure the workflow fails as early as possible. A DataSource has information on
    the driver that the data should be read with, and is responsible for initializing
    this driver.
    """

    model_config = ConfigDict(extra="forbid")

    _used: bool = PrivateAttr(default=False)
    _fallback_driver_read: ClassVar[str]
    _fallback_driver_write: ClassVar[str]

    name: str
    uri: str
    data_adapter: DataAdapterBase
    driver: BaseDriver
    uri_resolver: URIResolver = Field(default_factory=ConventionResolver)
    data_type: ClassVar[DataType]
    root: Optional[str] = Field(
        default=None, exclude=True
    )  # root is already in the catalog.
    version: Optional[Union[str, int, float]] = Field(default=None)
    provider: Optional[str] = Field(default=None)
    metadata: SourceMetadata = Field(default_factory=SourceMetadata)

    def summary(self) -> Dict[str, Any]:
        """Return a summary of the DataSource."""
        summ: Dict[str, Any] = self.model_dump(include={"uri"})
        summ.update(
            {
                "data_type": self.__class__.data_type,
                "driver": self.driver.__repr_name__(),
                **self.metadata.model_dump(exclude_unset=True),
            }
        )
        return summ

    def _mark_as_used(self):
        """Mark the data adapter as used."""
        self._used = True

    def _log_start_read_data(self):
        """Log the start of the read data process."""
        logger.info(f"Reading {self.name} {self.data_type} data from {self.full_uri}")

    @model_validator(mode="before")
    @classmethod
    def _validate_data_type(cls, data: Any) -> Any:
        """Pydantic does not check class variables, so it is checked here."""
        if isinstance(data, dict):
            copy_data: dict = deepcopy(data)
            if data_type := copy_data.pop("data_type", None):
                if data_type != cls.data_type:
                    raise ValueError(f"'data_type' must be '{cls.data_type}'.")
            if not copy_data.get("driver"):
                copy_data["driver"] = cls._infer_default_driver(copy_data.get("uri"))
        return copy_data

    @classmethod
    def _infer_default_driver(
        cls, uri: str | None = None, driver: type[BaseDriver] | None = None
    ) -> str:
        if uri is None:
            return cls._fallback_driver_read
        _, extension = splitext(uri)
        driver = driver if driver else BaseDriver
        return next(
            (
                driver.name
                for driver in driver.find_all_possible_types()
                if extension in driver.SUPPORTED_EXTENSIONS
            ),
            cls._fallback_driver_read,
        )

    @model_validator(mode="after")
    def _validate_fs_equal_if_not_set(self) -> DataSource:
        """
        Validate and change the filesystems.

        They have to be equal between driver and uri resolver if they are not set.
        They can be different, but only if set explicitly.
        """
        driver_fs_set = "filesystem" in self.driver.model_fields_set
        uri_res_fs_set = "filesystem" in self.uri_resolver.model_fields_set
        if driver_fs_set ^ uri_res_fs_set:
            if driver_fs_set:
                self.uri_resolver.filesystem = self.driver.filesystem
            else:
                self.driver.filesystem = self.uri_resolver.filesystem
        return self

    @property
    def full_uri(self) -> str:
        """Join root with uri."""
        uri_is_url: bool = _is_valid_url(self.uri)
        if uri_is_url:
            # uri is fully self-describing
            return self.uri
        elif self.root and _is_valid_url(self.root):
            # use '/' to connect url parts
            return f"{self.root.rstrip('/')}/{self.uri}"
        # Local file, make absolute
        return _abs_path(self.root, self.uri)

    @model_serializer(mode="wrap")
    def _serialize(self, nxt: SerializerFunctionWrapHandler) -> Dict[str, Any]:
        """Serialize data_type."""
        res: Dict[str, Any] = nxt(self)
        res["data_type"] = self.data_type
        return res

    def _get_uri_basename(self, handle_nodata: NoDataStrategy, **query_kwargs) -> str:
        if "{" in self.uri:
            # first resolve any placeholders
            uris: List[str] = self.uri_resolver.resolve(
                uri=self.full_uri,
                handle_nodata=handle_nodata,
                **query_kwargs,
            )

            # if multiple_uris, use the first one:
            if len(uris) > 0:
                uri: str = uris[0]
            else:
                raise NoDataException("!")
        else:
            uri: str = self.uri

        basename: Optional[str] = PurePath(uri).name
        if basename is None:
            raise ValueError(f"Failed to get basename of uri: {self.uri}")
        else:
            return basename


def _abs_path(root: Union[Path, str], rel_path: Union[Path, str]) -> str:
    path = Path(str(rel_path))
    if not path.is_absolute():
        if root is not None:
            rel_path = join(root, rel_path)
        path = Path(abspath(rel_path))
    return str(path)



"""the new model root class."""

from logging import FileHandler, Logger, getLogger
from os import makedirs, remove
from os.path import dirname, exists, isdir, join
from pathlib import Path
from shutil import SameFileError, copyfile
from typing import Optional

from hydromt._typing import ModeLike, ModelMode
from hydromt._typing.type_def import StrPath
from hydromt._utils.log import _add_filehandler

logger = getLogger(__name__)

__all__ = ["ModelRoot"]


class ModelRoot:
    """A class to handle model roots in a cross platform manner."""

    def __init__(
        self,
        path: StrPath,
        mode: ModeLike = "w",
    ):
        self.set(path, mode)

    def set(
        self,
        path: StrPath,
        mode: Optional[ModeLike] = None,
    ) -> Path:
        """Set the path of the root, and create the necessary loggers."""
        if hasattr(self, "path"):
            old_path = self.path
        else:
            old_path = None

        logger.info(f"setting root to {path}")

        self.path = Path(path)

        if mode:
            self._mode = ModelMode.from_str_or_mode(mode)

        if self.is_reading_mode():
            self._check_root_exists()

        if self.mode is not None:
            is_override = self.mode.is_override_mode()
        else:
            is_override = False

        self._update_logger_filehandler(old_path, is_override)
        return self.path

    def _assert_write_mode(self) -> None:
        if not self.mode.is_writing_mode():
            raise IOError("Model opened in read-only mode")

    def _assert_read_mode(self) -> None:
        if not self.mode.is_reading_mode():
            raise IOError("Model opened in write-only mode")

    @property
    def mode(self) -> ModelMode:
        """The mode of the model this object belongs to."""
        return self._mode

    @mode.setter
    def mode(self, mode: ModeLike) -> ModelMode:
        """Set the mode of the model."""
        self._mode: ModelMode = ModelMode.from_str_or_mode(mode)
        return self._mode

    def is_writing_mode(self) -> bool:
        """Test whether we are in writing mode or not."""
        return self._mode.is_writing_mode()

    def is_reading_mode(self) -> bool:
        """Test whether we are in reading mode or not."""
        return self._mode.is_reading_mode()

    def is_override_mode(self) -> bool:
        """Test whether we are in override mode or not."""
        return self._mode.is_override_mode()

    def _update_logger_filehandler(
        self, old_path: Optional[Path] = None, overwrite: bool = False
    ):
        """Update the file handler to save logs in self.path/hydromt.log. If a log file at old_path exists copy it to self.path and append."""
        new_path = join(self.path, "hydromt.log")
        if old_path == new_path:
            return

        makedirs(self.path, exist_ok=True)

        main_logger: Logger = getLogger("hydromt")

        log_level = 20  # default, but overwritten by the level of active loggers
        for handler in main_logger.handlers:
            log_level = handler.level
            if isinstance(handler, FileHandler):
                if dirname(handler.baseFilename) != self.path:
                    # first remove so no new logs come in
                    main_logger.removeHandler(handler)
                    # wait for all messages to be processed
                    handler.flush()
                    # close the handler
                    handler.close()
                break

        if overwrite and exists(new_path):
            remove(new_path)

        if old_path is not None and exists(old_path):
            old_log_path = join(old_path, "hydromt.log")
            if exists(old_log_path):
                try:
                    copyfile(old_log_path, new_path)
                except SameFileError:
                    pass

        _add_filehandler(main_logger, new_path, log_level)

    def __repr__(self):
        return f"ModelRoot(path={self.path}, mode={self._mode})"

    def _check_root_exists(self):
        # check directory
        if not isdir(self.path):
            raise IOError(f'model root not found at "{self.path}"')



"""Base class for different drivers."""

from abc import ABC
from logging import Logger, getLogger
from typing import Any, ClassVar, Dict

from fsspec.implementations.local import LocalFileSystem
from pydantic import (
    ConfigDict,
    Field,
)

from hydromt._abstract_base import AbstractBaseModel
from hydromt._typing import FS
from hydromt.plugins import PLUGINS

logger: Logger = getLogger(__name__)


class BaseDriver(AbstractBaseModel, ABC):
    """Base class for different drivers.

    Is used to implement common functionality.
    """

    supports_writing: ClassVar[bool] = False
    filesystem: FS = Field(default_factory=LocalFileSystem)
    options: Dict[str, Any] = Field(default_factory=dict)

    model_config: ConfigDict = ConfigDict(extra="forbid")
    SUPPORTED_EXTENSIONS: ClassVar[set[str]] = set()

    def __eq__(self, value: object) -> bool:
        """Compare equality.

        Overwritten as filesystem between two sources can not be the same.
        """
        if isinstance(value, self.__class__):
            return self.model_dump(round_trip=True) == value.model_dump(round_trip=True)
        return super().__eq__(value)

    @classmethod
    def _load_plugins(cls):
        """Load Driver plugins."""
        plugins = PLUGINS.driver_plugins.keys()
        logger.debug(f"loaded {cls.__name__} plugins: {', '.join(plugins)}")



"""Abstract driver to read datasets."""

from abc import ABC, abstractmethod
from logging import Logger, getLogger
from typing import List, Optional

import xarray as xr

from hydromt._typing import (
    SourceMetadata,
    StrPath,
    TimeRange,
    Variables,
)
from hydromt.data_catalog.drivers.base_driver import BaseDriver
from hydromt.error import NoDataStrategy

logger: Logger = getLogger(__name__)


class DatasetDriver(BaseDriver, ABC):
    """Abstract Driver to read Datasets."""

    @abstractmethod
    def read(
        self,
        uris: List[str],
        *,
        variables: Optional[Variables] = None,
        time_range: Optional[TimeRange] = None,
        metadata: Optional[SourceMetadata] = None,
        handle_nodata: NoDataStrategy = NoDataStrategy.RAISE,
    ) -> xr.Dataset:
        """
        Read in any compatible data source to an xarray Dataset.

        args:
        """
        ...

    def write(
        self,
        path: StrPath,
        ds: xr.Dataset,
        **kwargs,
    ) -> str:
        """
        Write out a Dataset to file.

        Not all drivers should have a write function, so this method is not
        abstract.

        args:
        """
        raise NotImplementedError(
            f"Writing using driver '{self.name}' is not supported."
        )



"""Module for managing data catalogs and components, such as data sources, drivers, adapters and uri_resolvers."""

from .data_catalog import DataCatalog
from .predefined_catalog import PredefinedCatalog

__all__ = ["DataCatalog", "PredefinedCatalog"]
